#include "svm02.h"
#include "svm02_icodes.h"
#include <iostream>

void SVM02::handleInstruction(int32_t icode)
{
	if(icode >= 0)
	{
		checkStackFreeSpace(1);
		m_state.stack[m_state.sp++] = icode; //Non-negative values are simply pushed onto stack
	}
	else
	{
		switch(icode)
		{
		case IN: in(); break;
		case OUT: out(); break;
		case ADD: add(); break;
		case SUB: sub(); break;
		case MUL: mul(); break;
		case DIV: div(); break;
		case MOD: mod(); break;
		case NEG: neg(); break;
		case INC: inc(); break;
		case DEC: dec(); break;
		case AND: and_(); break;
		case OR: or_(); break;
		case NOT: not_(); break;
		case XOR: xor_(); break;
		case SHL: shl(); break;
		case SHR: shr(); break;
		case POP: pop(); break;
		case DUP: dup(); break;
		case SWP: swp(); break;
		case OVR: ovr(); break;
		case LOAD: load(); break;
		case STOR: stor(); break;
		case JMP: jmp(); break;
		case JE: je(); break;
		case JNE: jne(); break;
		case JG: jg(); break;
		case JGE: jge(); break;
		case JL: jl(); break;
		case JLE: jle(); break;
		case NOP: nop(); break;
		case HALT: halt(); break;
		case RF: rf(); break;
		case CRF: crf(); break;
		default: throw UnknownInstructionException(); break;
		}			
	}
}

void SVM02::in()
{
	checkStackFreeSpace(1);		
	long a;
	if(m_state.rf)
	{
		char c; 
		std::cin.read(&c, 1);
		a = static_cast<long>(c);
	}
	else
	{
		std::cout<<"?";
		std::cin >> a;
	}
	if(std::cin.bad())
	{
		throw VMException();
	}
	m_state.stack[m_state.sp++] = a;
}

void SVM02::out()
{
	checkStackSize(1);
	long a = m_state.stack[--m_state.sp];
	if(!m_state.rf)
	{
		std::cout << a << std::endl;
	}
	else
	{
		std::cout << static_cast<char>(a % 256);
	}
}

void SVM02::add()
{
	checkStackSize(2);
	long a = m_state.stack[m_state.sp-1] + m_state.stack[m_state.sp-2];
	m_state.stack[(m_state.sp--) - 2] = a;
}

void SVM02::sub()
{
	checkStackSize(2);
	long a = m_state.stack[m_state.sp-2] - m_state.stack[m_state.sp-1];
	m_state.stack[(m_state.sp--) - 2] = a;
}

void SVM02::mul()
{
	checkStackSize(2);
	long a = m_state.stack[m_state.sp-1] * m_state.stack[m_state.sp-2];
	m_state.stack[(m_state.sp--) - 2] = a;
}

void SVM02::div()
{
	checkStackSize(2);
	long a = m_state.stack[m_state.sp-2] / m_state.stack[m_state.sp-1];
	m_state.stack[(m_state.sp--) - 2] = a;
}

void SVM02::mod()
{
	checkStackSize(2);
	long a = m_state.stack[m_state.sp-2] % m_state.stack[m_state.sp-1];
	m_state.stack[(m_state.sp--) - 2] = a;
}

void SVM02::neg()
{
	checkStackSize(1);
	m_state.stack[m_state.sp-1] = -m_state.stack[m_state.sp-1];
}

void SVM02::inc()
{
	checkStackSize(1);
	m_state.stack[m_state.sp-1]++;
}

void SVM02::dec()
{
	checkStackSize(1);
	m_state.stack[m_state.sp-1]--;
}

void SVM02::and_()
{
	checkStackSize(2);
	long a = m_state.stack[m_state.sp-1] & m_state.stack[m_state.sp-2];
	m_state.stack[(m_state.sp--) - 2] = a;
}

void SVM02::or_()
{
	checkStackSize(2);
	long a = m_state.stack[m_state.sp-1] | m_state.stack[m_state.sp-2];
	m_state.stack[(m_state.sp--) - 2] = a;
}

void SVM02::xor_()
{
	checkStackSize(2);
	long a = m_state.stack[m_state.sp-1] ^ m_state.stack[m_state.sp-2];
	m_state.stack[(m_state.sp--) - 2] = a;
}

void SVM02::not_()
{
	checkStackSize(1);
	m_state.stack[m_state.sp-1] = !m_state.stack[m_state.sp-1];
}

void SVM02::shr()
{
	checkStackSize(2);
	long a = m_state.stack[m_state.sp-2] >> m_state.stack[m_state.sp-1];
	m_state.stack[(m_state.sp--) - 2] = a;
}

void SVM02::shl()
{
	checkStackSize(2);
	long a = m_state.stack[m_state.sp-2] << m_state.stack[m_state.sp-1];
	m_state.stack[(m_state.sp--) - 2] = a;
}

void SVM02::pop()
{
	checkStackSize(1);
	m_state.sp--;
}

void SVM02::dup()
{
	checkStackFreeSpace(1);
	checkStackSize(1);
	long a = m_state.stack[m_state.sp-1];
	m_state.stack[m_state.sp++] = a;
}

void SVM02::swp()
{
	checkStackSize(2);
	m_state.stack[m_state.sp-1] = m_state.stack[m_state.sp-1] ^ m_state.stack[m_state.sp-2];
	m_state.stack[m_state.sp-2] = m_state.stack[m_state.sp-1] ^ m_state.stack[m_state.sp-2];
	m_state.stack[m_state.sp-1] = m_state.stack[m_state.sp-1] ^ m_state.stack[m_state.sp-2];
}

void SVM02::ovr()
{
	checkStackSize(2);
	checkStackFreeSpace(1);
	m_state.stack[m_state.sp] = m_state.stack[m_state.sp - 2];
	++m_state.sp;
}

void SVM02::load()
{
	checkStackSize(1);
	checkMemoryAddress(m_state.stack[m_state.sp - 1]);
	m_state.stack[m_state.sp-1] = m_state.memory[m_state.stack[m_state.sp - 1]];
}

void SVM02::stor()
{
	checkStackSize(2);
	checkMemoryAddress(m_state.stack[m_state.sp - 1]);
	m_state.memory[m_state.stack[m_state.sp - 1]] = m_state.stack[m_state.sp - 2];
	m_state.sp -= 2;
}

void SVM02::jmp()
{
	checkStackSize(1);
	checkProgramMemoryAddress(m_state.stack[m_state.sp-1]);
	m_state.ip = m_state.stack[--m_state.sp] - 1;
}

void SVM02::je()
{
	checkStackSize(3);
	checkProgramMemoryAddress(m_state.stack[m_state.sp-1]);
	if(m_state.stack[m_state.sp-2] == m_state.stack[m_state.sp-3])
	{
		m_state.ip = m_state.stack[m_state.sp-1] - 1;
	}
	m_state.sp -= 3;
}

void SVM02::jne()
{
	checkStackSize(3);
	checkProgramMemoryAddress(m_state.stack[m_state.sp-1]);
	if(m_state.stack[m_state.sp-2] != m_state.stack[m_state.sp-3])
	{
		m_state.ip = m_state.stack[m_state.sp-1] - 1;
	}
	m_state.sp -= 3;
}

void SVM02::jg()
{
	checkStackSize(3);
	checkProgramMemoryAddress(m_state.stack[m_state.sp-1]);
	if(m_state.stack[m_state.sp-2] > m_state.stack[m_state.sp-3])
	{
		m_state.ip = m_state.stack[m_state.sp-1] - 1;
	}
	m_state.sp -= 3;
}

void SVM02::jge()
{
	checkStackSize(3);
	checkProgramMemoryAddress(m_state.stack[m_state.sp-1]);
	if(m_state.stack[m_state.sp-2] >= m_state.stack[m_state.sp-3])
	{
		m_state.ip = m_state.stack[m_state.sp-1] - 1;
	}
	m_state.sp -= 3;
}

void SVM02::jl()
{
	checkStackSize(3);
	checkProgramMemoryAddress(m_state.stack[m_state.sp-1]);
	if(m_state.stack[m_state.sp-2] < m_state.stack[m_state.sp-3])
	{
		m_state.ip = m_state.stack[m_state.sp-1] - 1;
	}
	m_state.sp -= 3;
}

void SVM02::jle()
{
	checkStackSize(3);
	checkProgramMemoryAddress(m_state.stack[m_state.sp-1]);
	if(m_state.stack[m_state.sp-2] <= m_state.stack[m_state.sp-3])
	{
		m_state.ip = m_state.stack[m_state.sp-1] - 1;
	}
	m_state.sp -= 3;
}

void SVM02::nop()
{
	//no-op
}

void SVM02::halt()
{
	m_state.hf = true;
}

void SVM02::rf()
{
	m_state.rf = true;
}

void SVM02::crf()
{
	m_state.rf = false;
}
