///@file compiler.h
///@author grigoryj
#ifndef COMPILER_H
#define COMPILER_H

#include "compiler_state.h"
#include "code_generator.h"
#include <iostream>
#include <vector>
#include <map>
#include <string>


///@class Compiler
///@brief Encapsulates details related to input parsing (via a finite state machine).
class Compiler
{
public:
	///@method Compiler
	///@brief Loads and initializes the instruction table
	Compiler();
	
public:
	///@method processInput
	///@brief Processes the program from input and generates "proto-code" which needs "finalization"
	///@param in The stream from which the program should be read
	void processInput(std::istream& in);

	///@method dumpCode
	///@brief Finalizes the compilation process, writes the resulting code to an output stream and resets the compiler's state
	///@param out The output stream
	void dumpCode(std::ostream& out);

private:
	///@method returnToInitialState
	///@brief Tries to return to initial state if possible.
	///@return True if returned to initial state
	bool returnToInitialState();
	
	///@brief Parses next line
	void parseNewLine();
	
	///@brief Parses a label's name
	void parseLabel();
	
	///@brief Parses an immediate integer value (positive or negative)
	void parseImmediateValue();
	
	///@brief Parses an instruction
	void parseInstruction();
	
	///@brief Parses a comment line
	void parseComment();
	
	///@brief Parses a character constant (characters can be escaped with \)
	void parseCharacterConstant();
	
	///@brief Parses a string constant
	void parseStringConstant();
	
private:
	static char getEscapeChar(char c);
		
private:
	CompilerState m_state;
	std::string m_line;
	char m_next_char;
	CodeGenerator m_codeGen;
};
#endif //COMPILER_H
